import asyncio
from .channel import Channel
from contextlib import asynccontextmanager, AsyncExitStack

class InputPort:
    def __init__(self, name, capacity=100, default=None):
        """
        Initializes an InputPort.
        
        :param name: The name of the InputPort.
        :param initialization_value: The initial value for the port, if any.
        """
        self.name = name
        self.component=None
        self.channel = Channel(capacity)  # Initialize a Channel object
        self.initialization_value = default

    def __repr__(self):
        return f'inport {self.component.name}.{self.name}'

    async def put(self, item):
        """
        Puts an item into the channel.
        
        :param item: The item to be put into the channel.
        """
        await self.channel.put(item)

    @asynccontextmanager
    async def open(self):
        """
        Context manager for managing the state of the channel.
        """
        async with self.channel.open() as channel:
            yield channel

    def initialize(self,value):
        """
            initialize this port with a value
        """
        self.initialization_value=value

    def __aiter__(self):
        """
        Makes InputPort an asynchronous iterable by delegating to the channel's iterator.
        """
        return self.channel.__aiter__()

    def __lt__(self,b):
        self.initialize(b)


class OutputPort:
    def __init__(self, name):
        """
        Initializes an OutputPort.
        
        :param name: The name of the OutputPort.
        """
        self.name = name
        self.component = None
        self.connections = set()  # List to store connections to InputPorts

    @asynccontextmanager
    async def open(self):
        async with AsyncExitStack() as stack:
            await asyncio.gather(*[stack.enter_async_context(port.open()) for port in self.connections])
            yield

    def __repr__(self):
        return f'inport {self.component.name}.{self.name}'

    async def send(self, item):
        """
        Concurrently sends an item to all connected InputPorts.
        
        :param item: The item to be sent.
        """
        await asyncio.gather(*(connection.put(item) for connection in self.connections))

    def connect(self, inport):
        """
        Connects this OutputPort to an InputPort.
        
        :param inport: The InputPort to connect to.
        """
        inport.initialization_value=None
        self.component.graph.edge(self, inport)
        if type(inport) is InputPort:
            self.connections.add(inport)

    def __rshift__(self,inport):
        self.connect(inport)


class PortHandler:
    def __init__(self, name, parent):
        self.name = name
        self.component = parent
    
    def __lt__(self,b):
        self.component.initialize(self.name, b)

    def __rshift__(self,inport):
        self.component.connect(self, inport)





import asyncio
import pickle
import multiprocessing
import time
from contextlib import asynccontextmanager
from multiprocessing import Manager

class Channel:

    def __init__(self, capacity):
        mgr = Manager()
        self._items = mgr.list()
        self.sender_count = multiprocessing.Value('i', 0)
        self._start_event = multiprocessing.Event()
        self._close_event = multiprocessing.Event()
        self._item_event = multiprocessing.Event()
        self._buf = []
        self._idx = 0

    @asynccontextmanager
    async def open(self):
        with self.sender_count.get_lock():
            self.sender_count.value += 1
            if self.sender_count.value == 1:
                self._start_event.set()
        try:
            yield self
        finally:
            with self.sender_count.get_lock():
                self.sender_count.value -= 1
                if self.sender_count.value == 0:
                    self._close_event.set()
                    self._item_event.set()

    async def put(self, item):
        self._items.append(item)
        self._item_event.set()

    def __aiter__(self):
        self._buf = []
        self._idx = 0
        return self

    async def __anext__(self):
        loop = asyncio.get_running_loop()
        if not self._start_event.is_set():
            await loop.run_in_executor(None, self._start_event.wait)
        if self._buf:
            return self._buf.pop(0)
        new_items = self._items[self._idx:]
        if new_items:
            self._idx += len(new_items)
            self._buf = new_items
            return self._buf.pop(0)
        if self._close_event.is_set():
            raise StopAsyncIteration
        self._item_event.clear()
        async def _wait():
            while True:
                if self._item_event.is_set() or self._close_event.is_set():
                    return
                await asyncio.sleep(0.001)
        await _wait()
        return await self.__anext__()



import asyncio
from collections import OrderedDict
from contextlib import AsyncExitStack
from aioitertools import zip as azip, chain, zip_longest, cycle
from copy import deepcopy
from .channel import Channel
from .port import InputPort, OutputPort, PortHandler




class ComponentMeta(type):
    def __mul__(cls, count):
        if not issubclass(cls, Component) or not isinstance(count, int):
            return NotImplemented
        return ComponentGroup(cls, count)

class Component(metaclass = ComponentMeta):
    def __init__(self, name):
        self.name = name
        self.group_name = None
        self.ports={}
        self.graph=None
        self.inports = set()
        self.outports = set()
        self._vals = OrderedDict()  # Holds references to both inports and outports
        self._runner = None
        self._groups = []
   
    def __repr__(self):
        inports_text = ' | '.join(port.name for port in self.inports)
        outports_text = ' | '.join(port.name for port in self.outports)
        component_text = self.name
        
        # Determining the longest string for centering
        max_length = max(len(inports_text), len(component_text), len(outports_text))
        
        # Centering each line based on the longest string
        inports_text_centered = inports_text.center(max_length)
        component_text_centered = component_text.center(max_length)
        outports_text_centered = outports_text.center(max_length)
        
        return f"\n{'_'*(max_length+4)}\n| {inports_text_centered} |\n| {component_text_centered} |\n| {outports_text_centered} |\n{'¯'*(max_length+4)}"

    def __getattr__(self,attr):
        return self.ports.get(attr)

    def copy(self):

        new = self.__class__(self.name)

        for ip in self.inports:
            new_ip = getattr(new, ip.name)
            new_ip.initialization_value = ip.initialization_value

        new._groups = [
            tuple(getattr(new, port.name) for port in group)
            for group in self._groups
        ]

        return new

    def add_inport(self, port):
        self.inports.add(port)
        port.component=self
        self.ports[port.name]=port
        self._vals[port]=None

    def add_outport(self, port):
        self.outports.add(port)
        port.component=self
        self.ports[port.name]=port
        self._vals[port]=port

    def set_runner(self, runner_function):
        self._runner = runner_function

    def clk(self, *ports):
        for port in ports:
            self._groups.append((port,))

    def sync(self, *ports):
        self._groups.append((ports))

    async def _group_listener(self, group, execute_runner):
        if execute_runner:
            async def listener(group):
                async for values in azip(*group):
                    self._vals.update(zip(group, values))
                    await self._runner(*self._vals.values())
        else:
            async def listener(group):
                async for values in azip(*group):
                    self._vals.update(zip(group, values))
        
        await listener(group)

    async def run(self):

        async with AsyncExitStack() as stack:
            await asyncio.gather(*[stack.enter_async_context(port.open()) for port in self.outports])
            # Categorize ports based on initialization values
            initports = {port for port in self.inports if port.initialization_value is not None}
    
            # Initialize ports with initialization values
            for port in initports:
                if isinstance(port.initialization_value, OutputPort):
                    async for value in port:
                        self._vals[port] = value
                        break
                else:
                    self._vals[port] = port.initialization_value

            # Check if all ports are initports
            if initports == self.inports:
                # If all inports are initports, call the runner and exit
                await self._runner(*self._vals.values())
                return

            if not self._groups:
                self._groups.append([port for port in self.inports if port.initialization_value is None])
    
            all_grouped_ports = set().union(*self._groups)
            exec_flags=[True for port in self._groups]
            ungrouped_ports = self.inports-all_grouped_ports-initports
    
            # Add each ungrouped port as a group on its own
            for port in ungrouped_ports:
                self._groups.append({port})
                exec_flags.append(False)
            # Prepare and run all group listeners concurrently
            group_listeners = [self._group_listener(group,flag) for group,flag in zip(self._groups,exec_flags)]
            await asyncio.gather(*group_listeners)


def inport(port_name, capacity=10, default=None):
    def decorator(func):
        if not hasattr(func, '_inports'):
            func._inports = []
        func._inports.append((port_name, capacity, default))
        return func
    return decorator

def outport(port_name):
    def decorator(func):
        if not hasattr(func, '_outports'):
            func._outports = []
        func._outports.append(port_name)
        return func
    return decorator

def component(func):
    class ComponentSubclass(Component):
        def __init__(self, *args, **kwargs):
            super().__init__(func.__name__)
            for args in getattr(func, '_inports', []):
                self.add_inport(InputPort(*args))
            for port_name in getattr(func, '_outports', []):
                self.add_outport(OutputPort(port_name))
            self.set_runner(deepcopy(func))

    ComponentSubclass.__name__ = func.__name__
    return ComponentSubclass

class ComponentGroup:
    def __init__(self, comp_cls, count):
        self.comp_cls = comp_cls     # the Component subclass
        self.count    = count        # number of copies
        # these get filled in by Graph._add_component_group
        self.name     = None
        self.graph    = None
        self.group    = []           # list of Component instances
        self.ports    = {}

    def __repr__(self):
        return f"<ComponentGroup {self.name} x{self.count}>"

    def __getattr__(self, attr):
        if not self.ports:
            for k,v in self.group[0].ports.items():
                self.ports[k]=PortHandler(k, self)
        return self.ports[attr]

    def initialize(self, name, val):
        for comp in self.group:
            comp.ports[name].__lt__(val)

    def connect(self, src, target):
        """
        Connect this group’s output-port `name` into `target`.
        - If `target` is a ComponentGroup of the same size: do 1:1 parallel wiring.
        - Otherwise, insert a M→N GroupConnector under a synthetic name.
        """
        

        if isinstance(src.component, ComponentGroup):
            M = src.component.count
            src_group = src.component.group
        else:
            src_group = [src.component]
            M = 1

        if isinstance(target.component, ComponentGroup):
            N = target.component.count            
            dst_group = target.component.group
        else:
            dst_group = [target.component]
            N = 1

        # Case A: same-size group→group → straight 1:1
        if M == N:
            for srcg, dstg in zip(enumerate(self.group), enumerate(dst_group)):
                i,src_comp = srcg
                j,dst_comp = dstg
                src_comp.ports[src.name] >> dst_comp.ports[target.name]
            return

        # Case B/C: M→N with M != N → insert a bridge
        conn_name = f"{src.component.name}.{src.name}-to-{target.component.name}.{target.name}"
        # create connector component with base port name `name`
        bridge = GroupConnector(conn_name, target.name, M, N)
        # register it in the graph
        self.graph.node(conn_name, bridge)
        # 1) wire each of the M group members into bridge.in_{i}
        for i, src_comp in enumerate(src_group):
            src_comp.ports[src.name] >> bridge.ports[f"{target.name}_{i}_in"]

        # 2) wire each of the N bridge.out_{j} into the dst components
        for j, dst_comp in enumerate(dst_group):
            bridge.ports[f"{target.name}_{j}_out"] >> dst_comp.ports[target.name]


class GroupConnector(Component):
    def __init__(self,base,name,M,N):
        self.name = base
        self.port_name = name
        super().__init__(base)
        self.M = M
        self.N = N
        for i in range(M): self.add_inport(InputPort(f"{name}_{i}_in"))
        for j in range(N): self.add_outport(OutputPort(f"{name}_{j}_out"))

    def copy(self):
        return GroupConnector(self.name, self.port_name, self.M, self.N)

    async def run(self):
        async with AsyncExitStack() as stack:
            await asyncio.gather(*[stack.enter_async_context(port.open()) for port in self.outports])
            in_iters = [self.ports[f"{self.port_name}_{i}_in"].__aiter__() for i in range(self.M)]
            merged   = chain.from_iterable(zip_longest(*in_iters, fillvalue=StopAsyncIteration))
            dist     = cycle(self.outports)
            async for item in merged:
                if item is StopAsyncIteration: break
                thisport = await anext(dist)
                await thisport.send(item)



from setuptools import setup, find_packages

setup(
    name='cascata',
    version='0.1.5',
    author='Stuart MacLeod',
    author_email='orangootan@gmail.com',
    description='An asynchronous, multicore graph execution framework for Python.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/simacleod/Cascata',
    project_urls={
        'Bug Tracker': 'https://github.com/simacleod/Cascata/issues',
    },
    license='MIT',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Framework :: AsyncIO',
    ],
    python_requires='>=3.7',
    install_requires=[
        'multiprocess',
        'dill',
        'aioprocessing',
        'aioitertools',
        'networkx',
    ],
    include_package_data=True,
    zip_safe=False
)
